//-----------------------------------------------------------
//    read_input.cpp
//-----------------------------------------------------------
#include "input_output.h"
#include "sykes/multi_carray.h"
#include "sykes/numeric/kronecker_delta.h"
#include "sykes/material/solid_utility.h"
#include <vector>
#include <stdexcept>
#include <iomanip>

namespace fem_elastic3d{
    //-----------------------------------------------------------
    //    readinput
    //-----------------------------------------------------------
    void read_input(fem_app& _femapp, std::string const& _filename)
    {
        std::ifstream inp(_filename.c_str());
        
        if(!inp) throw std::runtime_error("fem_elastic3d::read_input()");
        bool flag_elem = false, flag_node = false,
            flag_boundary = false, flag_material = false;
        
        while(inp){
            std::string str1;
            std::vector<std::string> vstr1;
            getline_skip_comment(inp, str1);
            split_string(str1, vstr1);
            std::cout << "readinput: " << str1 << std::endl;
            if(!inp) break;
            
            if(vstr1.at(0) == "ELEMENT"){
                std::cout << "init element" << std::endl;
                initialize_element_array_from_istream(inp, _femapp.fem_space.get_element_array(), std::stoi(vstr1.at(1)));
                flag_elem = true;
            }else if(vstr1.at(0) == "NODE"){
                std::cout << "init node" << std::endl;
                initialize_node_array_from_istream(inp, _femapp.fem_space.get_node_array(), std::stoi(vstr1.at(1)));
                flag_node = true;
            }else if(vstr1.at(0) == "BOUNDARY1"){
                std::cout << "init boundary1" << std::endl;
                initialize_boundary_from_istream(inp, _femapp.fem_space.get_dirichlet_boundary_condition(), std::stoi(vstr1.at(1)));
                flag_boundary = true;
            }else if(vstr1.at(0) == "BOUNDARY2"){
                std::cout << "init boundary2" << std::endl;
                initialize_boundary_from_istream(inp, _femapp.fem_space.get_neumann_boundary_condition(), std::stoi(vstr1.at(1)));
            }else if(vstr1.at(0) == "MATERIAL"){
                std::cout << "init material" << std::endl;
                initialize_material_array_from_istream(inp, _femapp.fem_space.get_material_array(), std::stoi(vstr1.at(1)));
                flag_material = true;
            }
        }
        
        if(!(flag_elem && flag_node && flag_boundary && flag_material)){
            throw std::runtime_error("fem_elastic3d::read_input");
        }
    }
    
    //-----------------------------------------------------------
    //    initialize_element_array_from_istream
    //-----------------------------------------------------------
    std::istream& initialize_element_array_from_istream(
        std::istream& _ist,
        std::vector<fem_app::element_type>& _elarray,
        std::size_t _elsize)
    {
        std::size_t const element_node_size = fem_app::element_type::node_size;
        
        try
        {
            for(std::size_t i1 = 0; (i1 < _elsize)&&_ist; i1++){
                std::string str1;
                std::vector<std::string> vstr1;
                std::vector<std::size_t> num_array(element_node_size, 0);
                
                getline_skip_comment(_ist, str1);
                split_string(str1, vstr1);
                for(std::size_t i2 = 0; i2 < element_node_size; i2++){
                    num_array[i2] = std::stoi(vstr1.at(i2));
                }
                
                std::size_t mate_num = 0;
                
                fem_app::element_type elem1(num_array);
                elem1.set_material_num(mate_num);
                _elarray.push_back(elem1);
            }
        }
        
        catch(...)
        {
            throw std::runtime_error("fem_elastic3d::initialize_element_array_from_istream");
        }
        if(!_ist){
            throw std::runtime_error("fem_elastic3d::initialize_element_array_from_istream");
        }
        return _ist;
    }
    
    //-----------------------------------------------------------
    //    initialize_node_array_from_istream
    //-----------------------------------------------------------
    std::istream& initialize_node_array_from_istream
        (std::istream& _ist, std::vector<sykes::multi_carray<realnum, 3>>& _node_array, std::size_t _node_size)
    {
        static const std::size_t dim = 3;
        try
        {
            _node_array.reserve(_node_size);
            for(std::size_t i1 = 0; (i1 < _node_size)&&_ist; i1++){
                std::string str1;
                std::vector<std::string> vstr1;
                sykes::multi_carray<realnum, dim> num_array;
                getline_skip_comment(_ist, str1);
                split_string(str1, vstr1);
                for(std::size_t i2 = 0; i2 < dim; i2++){
                    num_array[i2] = std::stod(vstr1.at(i2));
                }
                _node_array.push_back(num_array);
            }
        }
        
        catch(...)
        {
            throw std::runtime_error
            ("fem_elastic3d::initialize_node_array_from_istream");
        }
        if(!_ist){
            throw std::runtime_error("fem_elastic3d::initialize_node_array_from_istream");
        }
        return _ist;
    }
    
    //-----------------------------------------------------------
    //    initialize_material_array_from_istream
    //-----------------------------------------------------------
    std::istream& initialize_material_array_from_istream(
        std::istream& _ist,
        std::vector<fem_app::material_type>& _mate_array,
        std::size_t _mate_size)
    {
        static const std::size_t dim = 3;
        try
        {
            _mate_array.reserve(_mate_size);
            for(std::size_t i1 = 0; (i1 < _mate_size)&&_ist; i1++){
                std::string str1;
                std::vector<std::string> vstr1;
                fem_app::material_type mate1;
                double E = 0.0, poi = 0.0, rho = 0.0;
                
                getline_skip_comment(_ist, str1);
                split_string(str1, vstr1);
                E = std::stod(vstr1.at(0));
                poi = std::stod(vstr1.at(1));
                rho = std::stod(vstr1.at(2));
                mate1.set_density(rho);
                
                sykes::multi_carray<realnum, dim, dim, dim, dim> Dtmp, B1 = sykes::numeric::get_Bmat<3>();
                double mu = 0.5 * E / (1.0 + poi);
                double lamb = poi * E /((1.0 + poi) * (1.0 - 2.0 * poi));
                for(std::size_t i0 = 0; i0 < dim; ++i0){
                for(std::size_t i1 = 0; i1 < dim; ++i1){
                for(std::size_t i2 = 0; i2 < dim; ++i2){
                for(std::size_t i3 = 0; i3 < dim; ++i3){
                    Dtmp[i0][i1][i2][i3] =
                        2.0 * mu * B1[i0][i1][i2][i3]
                        + lamb * sykes::numeric::kdelta(i0, i1) * sykes::numeric::kdelta(i2, i3);
                }}}}
                mate1.set_elastic_modulus(Dtmp);
                _mate_array.push_back(mate1);
            }
        }
        
        catch(...)
        {
            throw std::runtime_error("fem_elastic3d::initialize_material_array_from_istream");
        }
        if(!_ist){
            throw std::runtime_error("fem_elastic3d::initialize_material_array_from_istream");
        }
        return _ist;
    }
    
    //-----------------------------------------------------------
    //    initialize_boundary_from_istream
    //-----------------------------------------------------------
    std::istream& initialize_boundary_from_istream
        (std::istream& _ist, std::vector<sykes::fem::boundary::boundary_condition>& _bound, std::size_t _bound_size)
    {
        try
        {
            _bound.reserve(_bound_size);
            for(std::size_t i1 = 0; (i1 < _bound_size)&&_ist; i1++){
                std::string str1;
                std::vector<std::string> vstr1;
                sykes::fem::boundary::boundary_condition bound1;
                
                getline_skip_comment(_ist, str1);
                split_string(str1, vstr1);
                bound1.node_num = std::stoi(vstr1.at(0));
                bound1.axis = std::stoi(vstr1.at(1));
                bound1.value = std::stod(vstr1.at(2));
                
                _bound.push_back(bound1);
            }
        }
        
        catch(...)
        {
            throw std::runtime_error("fem_elastic3d::initialize_boundary_from_istream");
        }
        if(!_ist){
            throw std::runtime_error("fem_elastic3d::initialize_boundary_array_from_istream");
        }
        return _ist;
    }
    
    //-----------------------------------------------------------
    //    output_result
    //-----------------------------------------------------------
    void output_result(fem_app const& _femapp, std::string const& _filename)
    {
        using sykes::material::solid::calc_cauchy_stress;
        std::ofstream outp(_filename.c_str());
        
        outp.setf(std::ios::scientific);
        if(!outp) throw std::runtime_error("fem_elastic3d::output_result()");
        
        auto const& nodes = _femapp.fem_space.get_node_array();
        auto const& elements = _femapp.fem_space.get_element_array();
        auto const& materials = _femapp.fem_space.get_material_array();
        auto const& rvec = _femapp.fem_space.get_result_vector();
        
        outp << "# vtk DataFile Version 2.0\n";
        outp << "cavity test\n";
        outp << "ASCII\n";
        outp << "DATASET UNSTRUCTURED_GRID\n";
        outp << "POINTS " << nodes.size() << " float\n";
        
        for(std::size_t i0 = 0, sz = nodes.size(); i0 < sz; ++i0){
            outp << std::setw(12) << nodes[i0][0];
            outp << " ";
            outp << std::setw(12) << nodes[i0][1];
            outp << " ";
            outp << std::setw(12) << nodes[i0][2];
            outp << " ";
            outp << "\n";
        }
        
        outp << "CELLS " << elements.size() << " " << elements.size() * (4 + 1) << std::endl;
        for(int i0 = 0, sz = elements.size(); i0 < sz; ++i0){
            sykes::multi_carray<std::size_t, 4> node_nums = elements[i0].get_node_nums();
            outp << std::setw(6) << 4;
            outp << " ";
            outp << std::setw(6) << node_nums[0];
            outp << " ";
            outp << std::setw(6) << node_nums[1];
            outp << " ";
            outp << std::setw(6) << node_nums[2];
            outp << " ";
            outp << std::setw(6) << node_nums[3];
            outp << std::endl;
        }
        
        
        outp << "CELL_TYPES " << elements.size() << std::endl;
        for(std::size_t i0 = 0; i0 < elements.size(); ++i0){
            outp << "10\n";
        }
        
        outp << "POINT_DATA " << nodes.size() << std::endl;
        outp << "VECTORS displacement float" << std::endl;
        for(std::size_t i0 = 0, sz = nodes.size(); i0 < sz; ++i0){
            outp << rvec(i0 * 3);
            outp << " ";
            outp << rvec(i0 * 3 + 1);
            outp << " ";
            outp << rvec(i0 * 3 + 2);
            outp << "\n";
        }
        
        outp << "CELL_DATA " << elements.size() << std::endl;
        outp << "TENSORS stress float" << std::endl;
        for(int i0 = 0, sz = elements.size(); i0 < sz; ++i0){
            std::size_t matenum = elements[i0].get_material_num();
            auto const& mate1 = materials[matenum];
            sykes::multi_carray<std::size_t, 4> node_nums = elements[i0].get_node_nums();
            sykes::multi_carray<realnum, 4, 3> pos, disp;
            
            for(std::size_t i1 = 0; i1 < 4; i1++){
                for(std::size_t i2 = 0; i2 < 3; i2++){
                    pos[i1][i2] = nodes[node_nums[i1]][i2];
                    disp[i1][i2] = rvec(node_nums[i1] * 3 + i2);
            }}
            
            sykes::multi_carray<realnum, 3, 3>
                stress = calc_cauchy_stress<fem_app::element_type>(pos, disp, mate1.get_elastic_modulus());
            
            for(int i1 = 0; i1 < 3; ++i1){
                for(int i2 = 0; i2 < 3; ++i2){
                    outp << std::dec << std::setprecision(12) << stress[i1][i2];
                    if(i1 != 2 || i2 != 2) outp << " ";
            }}
            outp << std::endl;
        }
        
    }
    
    //---- get_line_skip_comments
    std::istream& getline_skip_comment(std::istream& inp1, std::string& str1)
    {
        while(inp1){
            getline(inp1, str1);
            if((str1[0] != '#')&&(!is_whitespaces(str1))) break;
        }
        return inp1;
    }
    
    //---- split_string
    void split_string(std::string const& _str,
        std::vector<std::string>& _splited,
        std::string const& _delimiters)
    {
        std::string::const_iterator it1 = _str.begin(), end1 = _str.end();
        _splited.clear();
        while(it1 != end1){
            std::string strtmp;
            while(_delimiters.find(*it1) != std::string::npos && (it1 != end1)) ++it1;
            if(*it1 == '-' && (it1 != end1)){
                strtmp.push_back(*it1);
                ++it1;
                while(*it1 == ' ') ++it1;
            }
            while(_delimiters.find(*it1) == std::string::npos && (it1 != end1)){
                strtmp.push_back(*it1);
                ++it1;
            }
            if(strtmp != "") _splited.push_back(strtmp);
        }
    }
    
    //---- is_whitespaces
    bool is_whitespaces(std::string const& _str)
    {
        bool ws = true;
        auto ite1 = _str.begin(), end1 = _str.end();
        while(ite1 != end1){
            ws &= (bool)isspace((int)*ite1);
            ++ite1;
        }
        return ws;
    }
}//---- namespace fem_elastic2d



