//-----------------------------------------------------------
//    read_input.cpp
//-----------------------------------------------------------
#include "input_output.h"
#include "sykes/multi_carray.h"
#include "sykes/numeric/kronecker_delta.h"
#include "sykes/value_type.h"
#include <vector>
#include <tuple>
#include <stdexcept>
#include <iomanip>

namespace test_stokes2d{
    using sykes::realnum;
    //-----------------------------------------------------------
    //    readinput
    //-----------------------------------------------------------
    void read_input(fem_app& _femapp, std::string const& _filename)
    {
        std::ifstream inp(_filename.c_str());
        
        if(!inp) throw std::runtime_error("test_stokes2d::read_input()");
        bool flag_elem = false, flag_node = false,
            flag_boundary = false, flag_material = false;
        
        while(inp){
            std::string str1;
            std::vector<std::string> vstr1;
            getline_skip_comment(inp, str1);
            split_string(str1, vstr1);
            std::cout << "readinput: " << str1 << std::endl;
            if(!inp) break;
            
            if(vstr1.at(0) == "ELEMENT"){
                std::cout << "init element" << std::endl;
                initialize_element_array_from_istream(inp, _femapp.fem_space.get_element_array(), std::stoi(vstr1.at(1)));
                flag_elem = true;
            }else if(vstr1.at(0) == "NODE"){
                std::cout << "init node" << std::endl;
                initialize_node_array_from_istream(inp, _femapp.fem_space.get_node_array(), std::stoi(vstr1.at(1)));
                flag_node = true;
            }else if(vstr1.at(0) == "BOUNDARY1"){
                std::cout << "init boundary1" << std::endl;
                initialize_boundary_from_istream(inp, _femapp.fem_space.get_dirichlet_boundary_condition(), std::stoi(vstr1.at(1)));
                flag_boundary = true;
            }else if(vstr1.at(0) == "BOUNDARY2"){
                std::cout << "init boundary2" << std::endl;
                initialize_boundary_from_istream(inp, _femapp.fem_space.get_neumann_boundary_condition(), std::stoi(vstr1.at(1)));
            }else if(vstr1.at(0) == "MATERIAL"){
                std::cout << "init material" << std::endl;
                initialize_material_array_from_istream(inp, _femapp.fem_space.get_material_array(), std::stoi(vstr1.at(1)));
                flag_material = true;
            }
        }
        
        if(!(flag_elem && flag_node && flag_boundary && flag_material)){
            throw std::runtime_error("test_stokes2d::read_input");
        }
    }
    
    //-----------------------------------------------------------
    //    initialize_element_array_from_istream
    //-----------------------------------------------------------
    std::istream& initialize_element_array_from_istream(
        std::istream& _ist,
        std::vector<fem_app::element_type>& _elarray,
        std::size_t _elsize)
    {
        std::size_t const element_node_size = 6;
        try
        {
            for(std::size_t i1 = 0; (i1 < _elsize)&&_ist; i1++){
                std::string str1;
                std::vector<std::string> vstr1;
                std::vector<std::size_t> num_array(element_node_size, 0);
                
                getline_skip_comment(_ist, str1);
                split_string(str1, vstr1);
                for(std::size_t i2 = 0; i2 < element_node_size; i2++){
                    num_array[i2] = std::stoi(vstr1.at(i2));
                }
                
                std::size_t mate_num = 0;
                
                sykes::fem::element::eltype_tri<2> elem1(num_array);
                num_array.pop_back();
                num_array.pop_back();
                num_array.pop_back();
                sykes::fem::element::eltype_tri<1> elem2(num_array);
                elem1.set_material_num(mate_num);
                _elarray.push_back(std::make_tuple(elem1, elem2));
            }
        }
        
        catch(...)
        {
            throw std::runtime_error("test_stokes2d::initialize_element_array_from_istream");
        }
        if(!_ist){
            throw std::runtime_error("test_stokes2d::initialize_element_array_from_istream");
        }
        return _ist;
    }
    
    //-----------------------------------------------------------
    //    initialize_node_array_from_istream
    //-----------------------------------------------------------
    std::istream& initialize_node_array_from_istream
        (std::istream& _ist, std::vector<sykes::multi_carray<realnum, 2>>& _node_array, std::size_t _node_size)
    {
        static const std::size_t dim = 2;
        try
        {
            _node_array.reserve(_node_size);
            for(std::size_t i1 = 0; (i1 < _node_size)&&_ist; i1++){
                std::string str1;
                std::vector<std::string> vstr1;
                sykes::multi_carray<realnum, 2> num_array;
                getline_skip_comment(_ist, str1);
                split_string(str1, vstr1);
                for(std::size_t i2 = 0; i2 < dim; i2++){
                    num_array[i2] = std::stod(vstr1.at(i2));
                }
                _node_array.push_back(num_array);
            }
        }
        
        catch(...)
        {
            throw std::runtime_error
            ("stokes_flow2d::initialize_node_array_from_istream");
        }
        if(!_ist){
            throw std::runtime_error("stokes_flow2d::initialize_node_array_from_istream");
        }
        return _ist;
    }
    
    //-----------------------------------------------------------
    //    initialize_material_array_from_istream
    //-----------------------------------------------------------
    std::istream& initialize_material_array_from_istream(
        std::istream& _ist,
        std::vector<fem_app::material_type>& _mate_array,
        std::size_t _mate_size)
    {
        try
        {
            _mate_array.reserve(_mate_size);
            for(std::size_t i1 = 0; (i1 < _mate_size)&&_ist; i1++){
                std::string str1;
                std::vector<std::string> vstr1;
                fem_app::material_type mate1;
                
                getline_skip_comment(_ist, str1);
                split_string(str1, vstr1);
                realnum dtmp = std::stod(vstr1.at(0));
                realnum rho = std::stod(vstr1.at(1));
                mate1.set_density(rho);
                
                mate1.set_viscosity(dtmp);
                _mate_array.push_back(mate1);
            }
        }
        
        catch(...)
        {
            throw std::runtime_error("test_stokes2d::initialize_material_array_from_istream");
        }
        if(!_ist){
            throw std::runtime_error("test_stokes2d::initialize_material_array_from_istream");
        }
        return _ist;
    }
    
    //-----------------------------------------------------------
    //    initialize_boundary_from_istream
    //-----------------------------------------------------------
    std::istream& initialize_boundary_from_istream
        (std::istream& _ist, std::vector<sykes::fem::boundary::boundary_condition>& _bound, std::size_t _bound_size)
    {
        try
        {
            _bound.reserve(_bound_size);
            for(std::size_t i1 = 0; (i1 < _bound_size)&&_ist; i1++){
                std::string str1;
                std::vector<std::string> vstr1;
                sykes::fem::boundary::boundary_condition bound1;
                
                getline_skip_comment(_ist, str1);
                split_string(str1, vstr1);
                bound1.node_num = std::stoi(vstr1.at(0));
                bound1.axis = std::stoi(vstr1.at(1));
                bound1.value = std::stod(vstr1.at(2));
                
                _bound.push_back(bound1);
            }
        }
        
        catch(...)
        {
            throw std::runtime_error("test_stokes2d::initialize_boundary_from_istream");
        }
        if(!_ist){
            throw std::runtime_error("test_stokes2d::initialize_boundary_array_from_istream");
        }
        return _ist;
    }
    
    //-----------------------------------------------------------
    //    output_result
    //-----------------------------------------------------------
    void output_result(fem_app const& _femapp, std::string const& _filename)
    {
        std::ofstream outp(_filename.c_str());
        outp.setf(std::ios::scientific);
        if(!outp) throw std::runtime_error("test_stokes2d::output_result()");
        
        auto const& nodes = _femapp.fem_space.get_node_array();
        auto const& elements = _femapp.fem_space.get_element_array();
        auto const& rvec = _femapp.fem_space.get_result_vector();
        std::size_t node_size = nodes.size();
        
        outp << "# vtk DataFile Version 2.0\n";
        outp << "cavity test\n";
        outp << "ASCII\n";
        outp << "DATASET UNSTRUCTURED_GRID\n";
        outp << "POINTS " << node_size << " float\n";
        for(std::size_t i0 = 0; i0 < node_size; i0++){
            outp << nodes[i0][0];
            outp << " ";
            outp << nodes[i0][1];
            outp << " ";
            outp << 0.0;
            outp << "\n";
        }
        
        outp << "CELLS " << elements.size() << " " << elements.size() * (6 + 1) << std::endl;
        for(std::size_t i0 = 0; i0 < elements.size(); i0++){
            outp << 6;
            outp << " ";
            outp << std::get<0>(elements[i0]).get_node_num(0);
            outp << " ";
            outp << std::get<0>(elements[i0]).get_node_num(1);
            outp << " ";
            outp << std::get<0>(elements[i0]).get_node_num(2);
            outp << " ";
            outp << std::get<0>(elements[i0]).get_node_num(3);
            outp << " ";
            outp << std::get<0>(elements[i0]).get_node_num(4);
            outp << " ";
            outp << std::get<0>(elements[i0]).get_node_num(5);
            outp << " ";
            outp << "\n";
        }
        outp << "CELL_TYPES " << elements.size() << std::endl;
        for(std::size_t i0 = 0; i0 < elements.size(); i0++){
            outp << "22\n";
        }
        
        outp << "POINT_DATA " << node_size << std::endl;
        outp << "VECTORS velocity float" << std::endl;
        for(std::size_t i0 = 0, sz = nodes.size(); i0 < sz; i0++){
            outp << rvec(i0 * 3);
            outp << " ";
            outp << rvec(i0 * 3 + 1);
            outp << " ";
            outp << 0.0;
            outp << "\n";
        }
        
        outp << "CELL_DATA " << elements.size() << std::endl;
        outp << "SCALARS pressure float" << std::endl;
        outp << "LOOKUP_TABLE default" << std::endl;
        for(std::size_t i0 = 0, sz = elements.size(); i0 < sz; i0++){
            sykes::multi_carray<std::size_t, 3> node_nums = std::get<1>(elements[i0]).get_node_nums();
            double tmp = (
                rvec(node_nums[0] * 3 + 2)
                + rvec(node_nums[1] * 3 + 2)
                + rvec(node_nums[2] * 3 + 2)
                ) / 3.0;
            outp << tmp << std::endl;
        }
    }
    
    //---- get_line_skip_comments
    std::istream& getline_skip_comment(std::istream& inp1, std::string& str1)
    {
        while(inp1){
            getline(inp1, str1);
            if((str1[0] != '#')&&(!is_whitespaces(str1))) break;
        }
        return inp1;
    }
    
    //---- split_string
    void split_string(std::string const& _str,
        std::vector<std::string>& _splited,
        std::string const& _delimiters)
    {
        std::string::const_iterator it1 = _str.begin(), end1 = _str.end();
        _splited.clear();
        while(it1 != end1){
            std::string strtmp;
            while(_delimiters.find(*it1) != std::string::npos && (it1 != end1)) ++it1;
            if(*it1 == '-' && (it1 != end1)){
                strtmp.push_back(*it1);
                ++it1;
                while(*it1 == ' ') ++it1;
            }
            while(_delimiters.find(*it1) == std::string::npos && (it1 != end1)){
                strtmp.push_back(*it1);
                ++it1;
            }
            if(strtmp != "") _splited.push_back(strtmp);
        }
    }
    
    //---- is_whitespaces
    bool is_whitespaces(std::string const& _str)
    {
        bool ws = true;
        auto ite1 = _str.begin(), end1 = _str.end();
        while(ite1 != end1){
            ws &= (bool)isspace((int)*ite1);
            ++ite1;
        }
        return ws;
    }
}//---- namespace stokes_flow2d



