//-----------------------------------------------------------
//    equation_elastic_solid.h
//-----------------------------------------------------------
#ifndef SYKES_EQUATION_STEADY_LINEAR_ELASTIC_SOLID
#define SYKES_EQUATION_STEADY_LINEAR_ELASTIC_SOLID
#include "sykes/multi_carray.h"
#include "sykes/numeric/kronecker_delta.h"
#include "sykes/value_type.h"

namespace sykes{ namespace fem{ namespace equation{
namespace steady_linear_elastic_solid{
    //-----------------------------------------------------------
    //    calc_element_matrix
    //-----------------------------------------------------------
    template<typename Eltype>
    static multi_carray<realnum, Eltype::node_size, Eltype::node_size, Eltype::dim, Eltype::dim>
    calc_element_matrix(
        multi_carray<realnum, Eltype::node_size, Eltype::dim> const& _nodepos,
        multi_carray<realnum, Eltype::dim, Eltype::dim, Eltype::dim, Eltype::dim> const& _D)
    {
        multi_carray<realnum, Eltype::node_size, Eltype::node_size, Eltype::dim, Eltype::dim>
            IdNdxdNdx = element::calc_region_integral_dNdx_dNdx<Eltype>(_nodepos);
        multi_carray<realnum, Eltype::node_size, Eltype::node_size, Eltype::dim, Eltype::dim>
            emat;
    
        for(std::size_t n0 = 0; n0 < Eltype::node_size; ++n0){
        for(std::size_t n1 = 0; n1 < Eltype::node_size; ++n1){
        for(std::size_t i0 = 0; i0 < Eltype::dim; ++i0){
        for(std::size_t i1 = 0; i1 < Eltype::dim; ++i1){
            emat[n0][n1][i0][i1] = 0.0;
            for(std::size_t i2 = 0; i2 < Eltype::dim; ++i2){
            for(std::size_t i3 = 0; i3 < Eltype::dim; ++i3){
                emat[n0][n1][i0][i1] += _D[i0][i2][i3][i1] * IdNdxdNdx[n0][n1][i2][i3];
            }}
        }}}}
    
        return emat;
    }
    
    //-----------------------------------------------------------
    //    merge_element_matrix
    //-----------------------------------------------------------
    template<typename Eltype, typename Mat>
    static void merge_element_matrix(
        Mat& _mat,
        multi_carray<std::size_t, Eltype::node_size> const& _node_nums,
        multi_carray<realnum, Eltype::node_size, Eltype::node_size, Eltype::dim, Eltype::dim> const& _emat)
    {
        for(std::size_t n0 = 0; n0 < Eltype::node_size; ++n0){
        for(std::size_t n1 = 0; n1 < Eltype::node_size; ++n1){
        for(std::size_t i0 = 0; i0 < Eltype::dim; ++i0){
        for(std::size_t i1 = 0; i1 < Eltype::dim; ++i1){
            std::size_t x = _node_nums[n0] * Eltype::dim + i0;
            std::size_t y = _node_nums[n1] * Eltype::dim + i1;
            if(x >= y) _mat(x, y) += _emat[n0][n1][i0][i1];
        }}}}
    }
    //-----------------------------------------------------------
    //    merge_element_vector
    //-----------------------------------------------------------
    template<typename Eltype, typename Vec>
    static void merge_element_vector(
        Vec& _vec,
        multi_carray<std::size_t, Eltype::node_size> const& _node_nums,
        multi_carray<realnum, Eltype::node_size, Eltype::dim> const& _evec)
    {
        for(std::size_t n0 = 0; n0 < Eltype::node_size; ++n0){
        for(std::size_t i0 = 0; i0 < Eltype::dim; ++i0){
            std::size_t vpos = _node_nums[n0] * Eltype::dim + i0;
            _vec(vpos) += _evec[n0][i0];
        }}
    }

    //-----------------------------------------------------------
    //    compose_total_matrix_and_vector
    //-----------------------------------------------------------
    template<typename FemSpace>
    static void compose_total_matrix_and_vector(
        multi_carray<realnum, FemSpace::dim> _gravity,
        FemSpace& _fem_space)
    {
        typedef typename FemSpace::element_type element_type;
        static std::size_t const dim = FemSpace::dim;
        static std::size_t const elnode_size = element_type::node_size;
        
        auto const& elements = _fem_space.get_element_array();
        auto const& nodes = _fem_space.get_node_array();
        auto const& materials = _fem_space.get_material_array();
        auto& total_mat = _fem_space.get_total_matrix();
        auto& total_vec = _fem_space.get_total_vector();
        
        //---- total matrix, total vector
        for(std::size_t n0 = 0, sz = elements.size(); n0 < sz; ++n0){
            multi_carray<std::size_t, elnode_size> const& node_nums =
                elements[n0].get_node_nums();
            multi_carray<realnum, elnode_size, dim> nodepos;
            
            for(std::size_t i0 = 0; i0 < elnode_size; ++i0){
            for(std::size_t i1 = 0; i1 < dim; ++i1){
                nodepos[i0][i1] = nodes[node_nums[i0]][i1];
            }}
            
            std::size_t material_num = elements[n0].get_material_num();
            auto const& material1 = materials[material_num];
            realnum const density = material1.get_density();
            multi_carray<realnum, dim, dim, dim, dim> const&
                D1 = material1.get_elastic_modulus();
        
            //---- calc element matrix
            multi_carray<realnum, elnode_size, elnode_size, dim, dim>
                emat = calc_element_matrix<element_type>(nodepos, D1);
            merge_element_matrix<element_type>(total_mat, node_nums, emat);
        
            //---- calc element vector
            multi_carray<realnum, elnode_size>
                I_N = element::calc_region_integral_N<element_type>(nodepos);
            
            multi_carray<realnum, elnode_size, dim> evec;
            for(std::size_t i0 = 0; i0 < elnode_size; ++i0){
                for(std::size_t i1 = 0; i1 < dim; ++i1){
                    evec[i0][i1] = density * _gravity[i1] * I_N[i0];
            }}
            merge_element_vector<element_type>(total_vec, node_nums, evec);
        
        }
    }
}//---- namespace steady_linear_elastic_solid
}}}//---- namespace equation fem sykes
#endif

