//-----------------------------------------------------------
//    fluid_stokes_flow.h
//-----------------------------------------------------------
#ifndef SYKES_EQUATION_STOKES_FLOW
#define SYKES_EQUATION_STOKES_FLOW

#include <functional>
#include <algorithm>
#include <type_traits>
#include <tuple>
#include <utility>
#include "sykes/multi_carray.h"
#include "sykes/numeric/kronecker_delta.h"
#include "sykes/value_type.h"

namespace sykes{ namespace fem{ namespace equation{
namespace steady_uncompressible_stokes_flow_mixed_interpolation{
    //-----------------------------------------------------------
    //    calc_uu_element_matrix
    //-----------------------------------------------------------
    template<typename Eltype1>
    static multi_carray<realnum, Eltype1::node_size, Eltype1::node_size, Eltype1::dim, Eltype1::dim>
    calc_uu_element_matrix(
        multi_carray<realnum, Eltype1::node_size, Eltype1::dim> const& _nodepos,
        realnum _visco)
    {
        multi_carray<realnum, Eltype1::dim, Eltype1::dim, Eltype1::dim, Eltype1::dim> const
            B = numeric::get_Bmat<Eltype1::dim>();
        multi_carray<realnum, Eltype1::node_size, Eltype1::node_size, Eltype1::dim, Eltype1::dim>
            IdNdxdNdx = element::calc_region_integral_dNdx_dNdx<Eltype1>(_nodepos);
        multi_carray<realnum, Eltype1::node_size, Eltype1::node_size, Eltype1::dim, Eltype1::dim>
            emat;
        realnum f1 = 2.0 * _visco;
    
        for(std::size_t n0 = 0; n0 < Eltype1::node_size; ++n0){
        for(std::size_t n1 = 0; n1 < Eltype1::node_size; ++n1){
        for(std::size_t i0 = 0; i0 < Eltype1::dim; ++i0){
        for(std::size_t i1 = 0; i1 < Eltype1::dim; ++i1){
            emat[n0][n1][i0][i1] = 0.0;
            for(std::size_t i2 = 0; i2 < Eltype1::dim; ++i2){
            for(std::size_t i3 = 0; i3 < Eltype1::dim; ++i3){
                emat[n0][n1][i0][i1] += B[i0][i2][i3][i1] * IdNdxdNdx[n0][n1][i2][i3];
            }}
            emat[n0][n1][i0][i1] *= f1;
        }}}}
    
        return emat;
    }
    
    //-----------------------------------------------------------
    //    calc_up_element_matrix
    //-----------------------------------------------------------
    template<typename Eltype1, typename Eltype2>
    static multi_carray<realnum, Eltype1::node_size, Eltype2::node_size, Eltype1::dim>
    calc_up_element_matrix(
        multi_carray<realnum, Eltype1::node_size, Eltype1::dim> const& _nodepos1,
        multi_carray<realnum, Eltype2::node_size, Eltype2::dim> const& _nodepos2)
    {
        multi_carray<realnum, Eltype1::node_size, Eltype2::node_size, Eltype1::dim>
            IdNdxL = element::calc_region_integral_dNdx_L<Eltype1, Eltype2>(_nodepos1, _nodepos2);
    
        std::transform(IdNdxL.begin(), IdNdxL.end(), IdNdxL.begin(), std::negate<realnum>());
    
        return IdNdxL;
    }
    
    //-----------------------------------------------------------
    //    merge_element_matrices
    //-----------------------------------------------------------
    template<typename Eltype1, typename Eltype2, typename Mat>
    static void merge_element_matrices(
        Mat& _mat, multi_carray<std::size_t, Eltype1::node_size> const& _vmesh_node_nums,
        multi_carray<std::size_t, Eltype2::node_size> const& _pmesh_node_nums,
        multi_carray<realnum, Eltype1::node_size, Eltype1::node_size, Eltype1::dim, Eltype1::dim> const& _uuemat,
        multi_carray<realnum, Eltype1::node_size, Eltype2::node_size, Eltype1::dim> const& _upemat)
    {
        static int const dim = Eltype1::dim;
        for(std::size_t n0 = 0; n0 < Eltype1::node_size; ++n0){
        for(std::size_t n1 = 0; n1 < Eltype1::node_size; ++n1){
            std::size_t x = (_vmesh_node_nums[n0]) * (Eltype1::dim + 1);
            std::size_t y = (_vmesh_node_nums[n1]) * (Eltype1::dim + 1);
            for(std::size_t i0 = 0; i0 < Eltype1::dim; ++i0){
            for(std::size_t i1 = 0; i1 < Eltype1::dim; ++i1){
                if(x + i0 >= y + i1)
                    _mat(x + i0, y + i1) += _uuemat[n0][n1][i0][i1];
            }}
            _mat(x + dim, x + dim) = 1.0;
            _mat(y + dim, y + dim) = 1.0;
        }}
        for(std::size_t n0 = 0; n0 < Eltype1::node_size; ++n0){
        for(std::size_t n1 = 0; n1 < Eltype2::node_size; ++n1){
            std::size_t x = (_vmesh_node_nums[n0]) * (Eltype1::dim + 1);
            std::size_t y = (_pmesh_node_nums[n1]) * (Eltype1::dim + 1) + dim;
            for(std::size_t i0 = 0; i0 < Eltype1::dim; ++i0){
                _mat(x + i0, y) += _upemat[n0][n1][i0];
            }
            _mat(y, y) = 0.0;
        }}
    }

    //-----------------------------------------------------------
    //    merge_element_vector
    //-----------------------------------------------------------
    template<typename Eltype1, typename Vec>
    static void merge_element_vector(
        Vec& _vec,
        multi_carray<std::size_t, Eltype1::node_size> const& _node_nums,
        multi_carray<realnum, Eltype1::node_size, Eltype1::dim> const& _evec)
    {
        for(std::size_t n0 = 0; n0 < Eltype1::node_size; ++n0){
        for(std::size_t i0 = 0; i0 < Eltype1::dim; ++i0){
            std::size_t vpos = _node_nums[n0] * (Eltype1::dim + 1) + i0;
            _vec(vpos) += _evec[n0][i0];
        }}
    }
    
    //-----------------------------------------------------------
    //    compose_total_matrix_and_vector
    //-----------------------------------------------------------
    template<typename FemSpace>
    static void compose_total_matrix_and_vector(
        multi_carray<realnum, FemSpace::dim> _gravity,
        FemSpace& _fem_space)
    {
        typedef typename std::tuple_element<0, typename FemSpace::element_type>::type element_type1;
        typedef typename std::tuple_element<1, typename FemSpace::element_type>::type element_type2;
        static_assert(((element_type1::order) > (element_type2::order)),
            "sykes::fem::equation::steady_uncompressible_stokes_flow_mixed_interpolation::compose_total_matrix_and_vector()");
        
        static std::size_t const dim = FemSpace::dim;
        static std::size_t const elnode_size1 = element_type1::node_size;
        static std::size_t const elnode_size2 = element_type2::node_size;
        
        auto const& elements = _fem_space.get_element_array();
        auto const& nodes = _fem_space.get_node_array();
        auto const& materials = _fem_space.get_material_array();
        auto& total_mat = _fem_space.get_total_matrix();
        auto& total_vec = _fem_space.get_total_vector();
        
        //---- total matrix, total vector
        for(std::size_t n0 = 0, sz = elements.size(); n0 < sz; ++n0){
            multi_carray<std::size_t, elnode_size1>
                vnode_nums = std::get<0>(elements[n0]).get_node_nums();
            multi_carray<std::size_t, elnode_size2>
                pnode_nums = std::get<1>(elements[n0]).get_node_nums();
            multi_carray<realnum, elnode_size1, dim> nodepos1;
            multi_carray<realnum, elnode_size2, dim> nodepos2;
            
            for(std::size_t i0 = 0; i0 < elnode_size1; ++i0){
            for(std::size_t i1 = 0; i1 < dim; ++i1){
                nodepos1[i0][i1] = nodes[vnode_nums[i0]][i1];
            }}
            for(std::size_t i0 = 0; i0 < elnode_size2; ++i0){
            for(std::size_t i1 = 0; i1 < dim; ++i1){
                nodepos2[i0][i1] = nodes[pnode_nums[i0]][i1];
            }}
            
            std::size_t material_num = std::get<0>(elements[n0]).get_material_num();
            auto const& material1 = materials[material_num];
            realnum const density = material1.get_density();
            realnum const visco = material1.get_viscosity();
            
            //---- calc element matrix
            multi_carray<realnum, elnode_size1, elnode_size1, dim, dim>
                uu_emat = calc_uu_element_matrix<element_type1>(nodepos1, visco);
        
            multi_carray<realnum, elnode_size1, elnode_size2, dim>
                up_emat = calc_up_element_matrix<element_type1, element_type2>(nodepos1, nodepos2);
            
            merge_element_matrices<element_type1, element_type2>(total_mat, vnode_nums, pnode_nums, uu_emat, up_emat);
        
            //---- calc element vector
            multi_carray<realnum, elnode_size1>
                I_N = element::calc_region_integral_N<element_type1>(nodepos1);
        
            multi_carray<realnum, elnode_size1, dim> evec;
            for(std::size_t i0 = 0; i0 < elnode_size1; i0++){
            for(std::size_t i1 = 0; i1 < dim; i1++){
                    evec[i0][i1] = density * _gravity[i1] * I_N[i0];
            }}
            merge_element_vector<element_type1>(total_vec, vnode_nums, evec);
        
        }
    }

}//---- namespace steady_uncompressible_stokes_flow_mixed_interpolation
}}}//---- namespace equation fem sykes

#endif

