//-----------------------------------------------------------
//    fem_space.h
//-----------------------------------------------------------
#ifndef SYKES_FEM_SPACE
#define SYKES_FEM_SPACE

#include <cstddef>
#include <vector>
#include <tuple>
#include <type_traits>
#include "sykes/multi_carray.h"
#include "sykes/value_type.h"
#include "sykes/boundary_condition.h"
#include "sykes/utility.h"

namespace sykes{ namespace fem{
    //-----------------------------------------------------------
    //    struct fem_space
    //-----------------------------------------------------------
    template<
        typename Eltype,
        typename Material,
        typename Mat,
        typename Vec>
    class fem_space
    {
    public:
        static std::size_t const dim = element_type_dimension<Eltype>::value;
        typedef Eltype element_type;
        typedef multi_carray<realnum, dim> node_type;
        typedef Mat matrix_type;
        typedef Vec vector_type;
        typedef Material material_type;
        
        //---- constructor
        fem_space()
            : m_total_matrix(), m_total_vector(), m_result(),
                m_element_array(), m_node_array(), m_material_array(),
                m_dirichlet_boundary(), m_neumann_boundary()
            {}
        
        fem_space(
            std::vector<element_type> const& _elem,
            std::vector<node_type> const& _node,
            std::vector<material_type> const& _mate,
            std::vector<boundary::boundary_condition> const& _diri,
            std::vector<boundary::boundary_condition> const& _neum)
            : m_total_matrix(), m_total_vector(), m_result(),
                m_element_array(_elem), m_node_array(_node), m_material_array(_mate),
                m_dirichlet_boundary(_diri), m_neumann_boundary(_neum)
            {}
        
        //---- destructor
        ~fem_space(){}
        
        //---- total matrix interface
        matrix_type const& get_total_matrix() const
            { return m_total_matrix; }
        matrix_type& get_total_matrix()
            { return m_total_matrix; }
        void set_total_matrix(matrix_type const& _mat)
            { m_total_matrix = _mat; }
        
        //---- total vector interface
        vector_type const& get_total_vector() const
            { return m_total_vector; }
        vector_type& get_total_vector()
            { return m_total_vector; }
        void set_total_vector(vector_type const& _vec)
            { m_total_vector = _vec; }
        
        //---- result vector interface
        vector_type const& get_result_vector() const
            { return m_result; }
        vector_type& get_result_vector()
            { return m_result; }
        void set_result_vector(vector_type const& _vec)
            { m_result = _vec; }
        
        //---- element_array interface
        std::vector<element_type> const& get_element_array() const
            { return m_element_array; }
        std::vector<element_type>& get_element_array()
            { return m_element_array; }
        void set_element_array(std::vector<element_type> const& _elem)
            { m_element_array = _elem; }
        
        //---- node_array interface
        std::vector<node_type> const& get_node_array() const
            { return m_node_array; }
        std::vector<node_type>& get_node_array()
            { return m_node_array; }
        void set_node_array(std::vector<material_type> const& _node)
        { m_node_array = _node; }
        
        //---- material_array interface
        std::vector<material_type> const& get_material_array() const
            { return m_material_array; }
        std::vector<material_type>& get_material_array()
            { return m_material_array; }
        void set_material_array(std::vector<material_type> const& _mate)
            { m_material_array = _mate; }
        
        //---- dirichlet boundary interface
        std::vector<boundary::boundary_condition> const& get_dirichlet_boundary_condition() const
            { return m_dirichlet_boundary; }
        std::vector<boundary::boundary_condition>& get_dirichlet_boundary_condition()
            { return m_dirichlet_boundary; }
        void set_dirichlet_boundary_condtion(std::vector<boundary::boundary_condition> const& _diri)
            { m_dirichlet_boundary = _diri; }
        
        //---- neumann boundary interface
        std::vector<boundary::boundary_condition> const& get_neumann_boundary_condition() const
            { return m_neumann_boundary; }
        std::vector<boundary::boundary_condition>& get_neumann_boundary_condition()
            { return m_neumann_boundary; }
        void set_neumann_boundary_condtion(std::vector<boundary::boundary_condition> const& _neum)
            { m_neumann_boundary = _neum; }
        
    private:
        matrix_type m_total_matrix;
        vector_type m_total_vector;
        vector_type m_result;
        
        std::vector<element_type> m_element_array;
        std::vector<node_type> m_node_array;
        std::vector<material_type> m_material_array;
        std::vector<boundary::boundary_condition> m_dirichlet_boundary;
        std::vector<boundary::boundary_condition> m_neumann_boundary;
    };
    
    template<std::size_t Dim>
    struct node_sorter;
    
    //-----------------------------------------------------------
    //    reduce_band_width
    //-----------------------------------------------------------
    template<typename FemSpace>
    typename std::enable_if<
        is_tuple<typename FemSpace::element_type>::value,
        void
    >::type
    reduce_band_width(FemSpace& _fem_space)
    {
        typedef typename std::tuple_element<0,
            typename FemSpace::element_type>::type element_type1;
        typedef typename std::tuple_element<1,
            typename FemSpace::element_type>::type element_type2;
        
        static std::size_t const dim = FemSpace::dim, elnode_size1 = element_type1::node_size,
            elnode_size2 = element_type2::node_size;
        auto& nodes = _fem_space.get_node_array();
        std::vector<std::tuple<std::size_t, sykes::multi_carray<double, dim>>> nodes_tmp;
        
        double max1[dim] = {0.0}, min1[dim] = {0.0}, range1[dim] = {0.0};
        for(std::size_t i = 0, sz = nodes.size(); i < sz; ++i){
            for(std::size_t j = 0; j < dim; ++j){
                max1[j] = std::max(max1[j], nodes[i][j]);
                min1[j] = std::min(min1[j], nodes[i][j]);
            }
            nodes_tmp.push_back(std::make_tuple(i, nodes[i]));
        }
        for(std::size_t j = 0; j < dim; ++j) range1[j] = max1[j] - min1[j];
        
        multi_carray<std::size_t, dim> axis1;
        axis1[dim - 1] = std::max_element(&range1[0], &range1[0] + dim) - &range1[0];
        axis1[0] = std::min_element(&range1[0], &range1[0] + dim) - &range1[0];
        if(axis1[0] == axis1[dim - 1]) ++axis1[0];
        if(dim == 3) axis1[1] = 3 - axis1[0] - axis1[dim - 1];
        
        std::sort(nodes_tmp.begin(), nodes_tmp.end(), node_sorter<dim>(axis1));
        
        for(std::size_t i = 0, sz = nodes.size(); i < sz; ++i)
            nodes[i] = std::get<1>(nodes_tmp[i]);
        
        std::vector<std::size_t> nlist(nodes.size());
        for(std::size_t i = 0, sz = nodes.size(); i < sz; ++i)
            nlist[std::get<0>(nodes_tmp[i])] = i;
        
        auto& elements = _fem_space.get_element_array();
        for(std::size_t i = 0, sz = elements.size(); i < sz; ++i){
            auto& nodenums1 = std::get<0>(elements[i]).get_node_nums();
            auto& nodenums2 = std::get<1>(elements[i]).get_node_nums();
            for(std::size_t j = 0; j < elnode_size1; ++j)
                nodenums1[j] = nlist[nodenums1[j]];
            for(std::size_t j = 0; j < elnode_size2; ++j)
                nodenums2[j] = nlist[nodenums2[j]];
        }
        
        auto& boundary1 = _fem_space.get_dirichlet_boundary_condition();
        for(std::size_t i = 0, sz = boundary1.size(); i < sz; ++i)
            boundary1[i].node_num = nlist[boundary1[i].node_num];
        
        auto& boundary2 = _fem_space.get_neumann_boundary_condition();
        for(std::size_t i = 0, sz = boundary2.size(); i < sz; ++i)
            boundary2[i].node_num = nlist[boundary2[i].node_num];
    }
    
    //-----------------------------------------------------------
    //    reduce_band_width
    //-----------------------------------------------------------
    template<typename FemSpace>
    typename std::enable_if<
        !(is_tuple<typename FemSpace::element_type>::value),
        void
    >::type
    reduce_band_width(FemSpace& _fem_space)
    {
        typedef typename FemSpace::element_type element_type;
        
        static std::size_t const dim = FemSpace::dim, elnode_size = element_type::node_size;
        auto& nodes = _fem_space.get_node_array();
        std::vector<std::tuple<std::size_t, sykes::multi_carray<double, dim>>> nodes_tmp;
        
        double max1[dim] = {0.0}, min1[dim] = {0.0}, range1[dim] = {0.0};
        for(std::size_t i = 0, sz = nodes.size(); i < sz; ++i){
            for(std::size_t j = 0; j < dim; ++j){
                max1[j] = std::max(max1[j], nodes[i][j]);
                min1[j] = std::min(min1[j], nodes[i][j]);
            }
            nodes_tmp.push_back(std::make_tuple(i, nodes[i]));
        }
        for(std::size_t j = 0; j < dim; ++j) range1[j] = max1[j] - min1[j];
        
        multi_carray<std::size_t, dim> axis1;
        axis1[dim - 1] = std::max_element(&range1[0], &range1[0] + dim) - &range1[0];
        axis1[0] = std::min_element(&range1[0], &range1[0] + dim) - &range1[0];
        if(axis1[0] == axis1[dim - 1]) ++axis1[0];
        if(dim == 3) axis1[1] = 3 - axis1[0] - axis1[dim - 1];
        
        std::sort(nodes_tmp.begin(), nodes_tmp.end(), node_sorter<dim>(axis1));
        
        for(std::size_t i = 0, sz = nodes.size(); i < sz; ++i)
            nodes[i] = std::get<1>(nodes_tmp[i]);
        
        std::vector<std::size_t> nlist(nodes.size());
        for(std::size_t i = 0, sz = nodes.size(); i < sz; ++i)
            nlist[std::get<0>(nodes_tmp[i])] = i;
        
        auto& elements = _fem_space.get_element_array();
        for(std::size_t i = 0, sz = elements.size(); i < sz; ++i){
            auto& nodenums = elements[i].get_node_nums();
            for(std::size_t j = 0; j < elnode_size; ++j)
                nodenums[j] = nlist[nodenums[j]];
        }
        
        auto& boundary1 = _fem_space.get_dirichlet_boundary_condition();
        for(std::size_t i = 0, sz = boundary1.size(); i < sz; ++i)
            boundary1[i].node_num = nlist[boundary1[i].node_num];
        
        auto& boundary2 = _fem_space.get_neumann_boundary_condition();
        for(std::size_t i = 0, sz = boundary2.size(); i < sz; ++i)
            boundary2[i].node_num = nlist[boundary2[i].node_num];
    }
    
    //-----------------------------------------------------------
    //    node_sorter
    //-----------------------------------------------------------
    template<std::size_t Dim>
    class node_sorter
        : std::binary_function<
            std::tuple<std::size_t, multi_carray<realnum, Dim>>,
            std::tuple<std::size_t, multi_carray<realnum, Dim>>, bool>
    {
    public:
        typedef std::tuple<std::size_t, multi_carray<realnum, Dim>> node_type;
        
        node_sorter(multi_carray<std::size_t, Dim> const& a1)
            : axis(a1) {}
        
        bool operator()(node_type const& _lhs, node_type const& _rhs) const
        {
            for(std::size_t i = 0; i < Dim; ++i){
                if((std::get<1>(_lhs)[axis[i]]) < (std::get<1>(_rhs)[axis[i]]))
                    return true;
                else if((std::get<1>(_lhs)[axis[i]]) > (std::get<1>(_rhs)[axis[i]]))
                    return false;
            }
            return true;
        }
    private:
        multi_carray<std::size_t, Dim> axis;
    };
    
}}//---- namespace fem sykes
#endif

