//-----------------------------------------------------------
//    mixed_material.h
//-----------------------------------------------------------
#ifndef SYKES_MATERIAL_MIXED_MATERIAL
#define SYKES_MATERIAL_MIXED_MATERIAL

#include "sykes/multi_carray.h"
#include "sykes/utility.h"
#include "sykes/value_type.h"
#include <tuple>
#include <cstddef>
#include <type_traits>

namespace sykes{ namespace material{
    
    //-----------------------------------------------------------
    //    struct mixed_material
    //-----------------------------------------------------------
    template<typename ... Material>
    class mixed_material;
    
    template<typename Mate1, typename Mate2, typename ... Mate3>
    class mixed_material<Mate1, Mate2, Mate3...> :
        public Mate1, public mixed_material<Mate2, Mate3...>
    {
        static_assert(!std::is_base_of<Mate1, mixed_material<Mate2, Mate3...>>::value,
            "class mixed_material");
    public:
        typedef std::tuple<Mate1, Mate2, Mate3...> material_types;
        typedef Mate1 base_type1;
        typedef mixed_material<Mate2, Mate3...> base_type2;
        
        mixed_material(){}
        
        template<typename ... Types>
        mixed_material(
            typename Mate1::property_type _a1,
            Types ... _args)
                : base_type1(_a1), base_type2(_args...)
        {}
        
    };
    
    template<typename Mate1>
    class mixed_material<Mate1> : public Mate1
    {
    public:
        typedef std::tuple<Mate1> material_types;
        typedef Mate1 base_type1;
        
        mixed_material() : Mate1() {}
        
        mixed_material(typename base_type1::property_type _a1)
            : base_type1(_a1)
        {}
    };
    
    //-----------------------------------------------------------
    //    scalar
    //-----------------------------------------------------------
    template<std::size_t ID>
    class scalar_property
    {
    public:
        static std::size_t const identity = ID;
        typedef realnum property_type;
        scalar_property() : m_scalar_property() {}
        explicit scalar_property(property_type const& _val)
            : m_scalar_property(_val) {}
        ~scalar_property(){}
        
        property_type get()
        { return m_scalar_property; }
        void set(realnum _val)
        { m_scalar_property = _val; }
    
    private:
        property_type m_scalar_property;
    };
    
    //-----------------------------------------------------------
    //    vector
    //-----------------------------------------------------------
    template<std::size_t ID, std::size_t Dim>
    class vector_property
    {
    public:
        typedef multi_carray<realnum, Dim> property_type;
        static std::size_t const identity = ID;
        static std::size_t const dim = Dim;
        
        ~vector_property(){}
        
        property_type get()
        { return m_vector_property; }
        void set(realnum _val)
        { m_vector_property = _val; }
        
    private:
        multi_carray<realnum, Dim> m_vector_property;
    };
    
    //-----------------------------------------------------------
    //    tensor
    //-----------------------------------------------------------
    template<std::size_t ID, std::size_t Dim>
    class tensor_property
    {
    public:
        typedef multi_carray<realnum, Dim, Dim> property_type;
        static std::size_t const identity = ID;
        static std::size_t const dim = Dim;
        
        tensor_property() : m_tensor_property() {}
        explicit tensor_property(property_type const& _val)
            : m_tensor_property(_val) {}
        ~tensor_property(){}
        
        property_type get()
        { return m_tensor_property; }
        void set(realnum _val)
        { m_tensor_property = _val; }
    
    private:
        multi_carray<realnum, Dim, Dim> m_tensor_property;
    };
    
    //-----------------------------------------------------------
    //    poisson
    //-----------------------------------------------------------
    class poisson_equation_factor
    {
    public:
        typedef realnum property_type;
        poisson_equation_factor() : m_poisson_equation_factor() {}
        explicit poisson_equation_factor(property_type const& _val)
            : m_poisson_equation_factor(_val) {}
        ~poisson_equation_factor(){}
        
        realnum    get_poisson_equation_factor() const
        { return m_poisson_equation_factor; }
        
        void set_poisson_equation_factor(realnum _poi)
        { m_poisson_equation_factor = _poi; }
    
    private:
        realnum m_poisson_equation_factor;
    };
    
    //-----------------------------------------------------------
    //    density
    //-----------------------------------------------------------
    class density
    {
    public:
        typedef realnum property_type;
        density() : m_density() {}
        explicit density(property_type const& _val)
            : m_density(_val) {}
        ~density(){}
        
        realnum    get_density() const
        { return m_density; }
        
        void set_density(realnum _dens)
        { m_density = _dens; }
    
    private:
        realnum m_density;
    };
    
    //-----------------------------------------------------------
    //    viscosity
    //-----------------------------------------------------------
    class viscosity
    {
    public:
        typedef realnum property_type;
        viscosity() : m_viscosity() {}
        explicit viscosity(property_type const& _val)
            : m_viscosity(_val) {}
        ~viscosity(){}
        
        realnum    get_viscosity() const
        { return m_viscosity; }
        
        void set_viscosity
        (realnum _visco)
        { m_viscosity = _visco; }
    
    private:
        realnum m_viscosity;
    };
    
    //-----------------------------------------------------------
    //    linear_elasticity
    //-----------------------------------------------------------
    template<std::size_t Dim>
    class linear_elasticity
    {
    public:
        typedef multi_carray<realnum, Dim, Dim, Dim, Dim> property_type;
        static int const dim = Dim;
        linear_elasticity() : m_elastic_modulus() {}
        explicit linear_elasticity(property_type const& _val)
            : m_elastic_modulus(_val) {}
        ~linear_elasticity(){}
        
        property_type const&
        get_elastic_modulus() const
        { return m_elastic_modulus; }
        
        void set_elastic_modulus
        (property_type const& _D)
        { m_elastic_modulus = _D; }
        
    private:
        property_type m_elastic_modulus;
    };
    
}}//---- namespace material sykes
#endif

