//-----------------------------------------------------------
//    interval_iterator.h
//-----------------------------------------------------------
#ifndef SYKES_NUMERIC_INTERVAL_ITERATOR
#define SYKES_NUMERIC_INTERVAL_ITERATOR

#include <iterator>
#include <type_traits>
#include <cstddef>

namespace sykes{ namespace numeric{
    //-----------------------------------------------------------
    //    class interval_iterator
    //-----------------------------------------------------------
    template<typename Iterator, typename Container>
    class interval_iterator
    {
        static_assert(
            std::is_base_of<
                std::bidirectional_iterator_tag,
                typename std::iterator_traits<Iterator>::iterator_category
            >::value
            || std::is_same<
                std::bidirectional_iterator_tag,
                typename std::iterator_traits<Iterator>::iterator_category
            >::value,
            "class interval_iterator");
    public:
        typedef interval_iterator self_type;
        typedef typename std::iterator_traits<Iterator>::value_type
            value_type;
        typedef typename std::iterator_traits<Iterator>::difference_type
            difference_type;
        typedef typename std::iterator_traits<Iterator>::pointer
            pointer;
        typedef typename std::iterator_traits<Iterator>::reference
            reference;
        typedef std::bidirectional_iterator_tag iterator_category;
        typedef typename Container::size_type size_type;
        
        //---- constructor
        inline interval_iterator()
            : m_interval(1), m_current(0){}
        
        inline interval_iterator(size_type _intv, Iterator _pos)
            : m_interval(_intv), m_current(_pos) {}
        
        template<typename Ite>
        inline interval_iterator(
            interval_iterator<
                Ite,
                typename std::enable_if<
                    std::is_same<
                        typename std::iterator_traits<Ite>::pointer,
                        typename Container::pointer
                    >::value,
                Container>::type> const& _other)
                : m_interval(_other.get_interval()), m_current(_other.base()) {}
        
        //---- operators
        inline reference operator*() const
        { return *m_current; }
        
        inline self_type& operator++()
        {
            m_current += m_interval;
            return *this;
        }
    
        inline self_type operator++(int)
        {
            self_type tmp(*this);
            m_current += m_interval;
            return tmp;
        }
    
        inline self_type& operator--()
        {
            m_current -= m_interval;
            return *this;
        }
    
        inline self_type operator--(int)
        {
            self_type tmp(*this);
            m_current -= m_interval;
            return tmp;
        }
        
        inline bool operator==(self_type const& _other) const
        {
            return (m_current == _other.m_current);
        }
    
        inline bool operator!=(self_type const& _other) const
        {
            return (m_current != _other.m_current);
        }
        
        template<typename Ite>
        bool operator==(
            interval_iterator<
                Ite,
                typename std::enable_if<
                    std::is_same<
                        typename std::iterator_traits<Ite>::pointer,
                        typename Container::const_pointer
                    >::value,
                Container>::type> const& _other) const
        { return (base() == _other.base()); }
        
        template<typename Ite>
        bool operator!=(
            interval_iterator<
                Ite,
                typename std::enable_if<
                    std::is_same<
                        typename std::iterator_traits<Ite>::pointer,
                        typename Container::const_pointer
                    >::value,
                Container>::type> const& _other) const
        { return (base() != _other.base()); }
        
        //---- interval interface
        size_type get_interval() const
        { return m_interval; }
        
        void set_interval(size_type _intv)
        { m_interval = _intv; }
        
        //----
        Iterator base() const
        {
            return m_current;
        }
    private:
        size_type m_interval;
        Iterator m_current;
    };//---- class interval_iterator
    
    //---- iterator difference
    template<typename Iterator, typename Container>
    inline typename interval_iterator<Iterator, Container>::difference_type 
    operator-(interval_iterator<Iterator, Container> const& _lhs,
        interval_iterator<Iterator, Container> const& _rhs)
    {
        typedef typename interval_iterator<Iterator, Container>::difference_type
            difference_type;
        difference_type tmp = (_lhs.base() - _rhs.base()) / _lhs.get_interval();
        return tmp;
    }
    
    template<typename LI, typename RI, typename Container>
    inline typename interval_iterator<LI, Container>::difference_type 
    operator-(
        interval_iterator<LI, Container> const& _lhs,
        interval_iterator<
                RI,
                typename std::enable_if<
                    std::is_same<
                        typename std::iterator_traits<RI>::pointer,
                        typename Container::pointer
                    >::value
                    || std::is_same<
                        typename std::iterator_traits<RI>::pointer,
                        typename Container::const_pointer
                    >::value,
                Container>::type> const& _rhs)
    {
        typedef typename interval_iterator<LI, Container>::difference_type
            difference_type;
        difference_type tmp = (_lhs.base() - _rhs.base()) / _lhs.get_interval();
        return tmp;
    }
    
}}//---- namespace numeric sykes
#endif

